@@ -1,11 +1,10 @@
 // If file logging is enabled
+use std::cell::RefCell;
 use std::fs::OpenOptions;
 use std::io::prelude::*;
 
 use crate::{Request, Server};
 
-static mut LOGGER: Option<Logger> = None;
-
 /// Define Log Levels
 pub enum Level {
     /// Give lots of information on what's going on.
@@ -139,15 +138,10 @@ impl Logger {
     /// server.start();
     /// ```
     pub fn attach(server: &mut Server, logger: Logger) {
-        unsafe {
-            LOGGER = Some(logger);
-        }
-
-        server.every(Box::new(|req| {
-            unsafe {
-                LOGGER.as_mut().unwrap().log(req);
-            }
+        let logger = RefCell::new(logger);
 
+        server.every(Box::new(move |req| {
+            logger.borrow_mut().log(req);
             None
         }));
     }
