@@ -1,4 +1,5 @@
 // Import STD libraries
+use std::io;
 use std::io::Read;
 use std::io::Write;
 use std::net::IpAddr;
@@ -8,8 +9,17 @@ use std::net::TcpListener;
 use std::net::TcpStream;
 use std::str;
 
-// Import local files
+// Feature Imports
+#[cfg(feature = "panic_handler")]
+use std::panic;
 
+#[cfg(feature = "cookies")]
+use super::cookie::Cookie;
+
+#[cfg(feature = "thread_pool")]
+use super::threadpool::ThreadPool;
+
+// Import local files
 use super::header::{headers_to_string, Header};
 use super::method::Method;
 use super::query::Query;
@@ -43,6 +53,9 @@ pub struct Server {
     /// Really just for testing.
     run: bool,
 
+    /// Default response for internal server errors
+    #[cfg(feature = "panic_handler")]
+    error_handler: fn(Request) -> Response,
     /// Headers automatically added to every response.
     default_headers: Option<Vec<Header>>,
 }
@@ -83,9 +96,17 @@ impl Server {
             port,
             ip,
             routes: Vec::new(),
-            run: true,
-            default_headers: Some(vec![Header::new("Server", "afire")]),
             middleware: Vec::new(),
+            run: true,
+            #[cfg(feature = "panic_handler")]
+            error_handler: |_| {
+                Response::new(
+                    500,
+                    "Internal Server Error :/",
+                    vec![Header::new("Content-Type", "text/plain")],
+                )
+            },
+            default_headers: Some(vec![Header::new("Server", "afire")]),
         }
     }
 
@@ -122,13 +143,7 @@ impl Server {
             return;
         }
 
-        let listener = TcpListener::bind(SocketAddr::new(
-            IpAddr::V4(Ipv4Addr::new(
-                self.ip[0], self.ip[1], self.ip[2], self.ip[3],
-            )),
-            self.port,
-        ))
-        .unwrap();
+        let listener = init_listener(self.ip, self.port).unwrap();
 
         for event in listener.incoming() {
             // Read stream into buffer
@@ -139,21 +154,18 @@ impl Server {
             let mut res = self.handle_connection(&stream);
 
             // Add default headers to response
-            if self.default_headers.is_some() {
-                for header in self.default_headers.as_ref().unwrap() {
-                    res.headers.push(header.copy());
-                }
-            }
+            let mut headers = res.headers;
+            headers.append(&mut self.default_headers.clone().unwrap_or_default());
 
             // Add content-length header to response
-            res.headers
-                .push(Header::new("Content-Length", &res.data.len().to_string()));
+            headers.push(Header::new("Content-Length", &res.data.len().to_string()));
 
             // Convert the response to a string
+            // TODO: Use Bytes instead of String
             let mut response = format!(
                 "HTTP/1.1 {} OK\r\n{}\r\n\r\n",
                 res.status,
-                headers_to_string(res.headers)
+                headers_to_string(headers)
             )
             .as_bytes()
             .to_vec();
@@ -167,79 +179,83 @@ impl Server {
         }
     }
 
-    /// Handel a connection to the server
-    // TODO: Try just expanding buffer if full and not relinging on content length
-    fn handle_connection(&self, mut stream: &TcpStream) -> Response {
-        // Init (first) Buffer
-        let mut buffer = vec![0; BUFF_SIZE];
-
-        // Read stream into buffer
-        let _ = stream.read(&mut buffer).unwrap();
-
-        // Get buffer as string
-        let buffer_clone = buffer.clone();
-        let stream_string = str::from_utf8(&buffer_clone).expect("Error parsing buffer data");
-
-        // Get Content-Length header
-        // If header shows thar more space is needed,
-        // make a new buffer read the rest of the stream and add it to the first buffer
-        for i in get_request_headers(stream_string.to_string()) {
-            if i.name != "Content-Length" {
-                continue;
-            }
-            let header_size = get_header_size(stream_string.to_string());
-            let content_length = i.value.parse::<usize>().unwrap_or(0);
-            let new_buffer_size = content_length as i64 + header_size as i64 - BUFF_SIZE as i64;
-            if new_buffer_size > 0 {
-                let mut new_buffer = vec![0; new_buffer_size as usize];
-                let _ = stream.read(&mut new_buffer).unwrap();
-                buffer.append(&mut new_buffer);
-            }
-            break;
+    /// Start the server with a thread pool.
+    /// ## Example
+    /// ```rust
+    /// // Import Library
+    /// use afire::{Server, Response, Header, Method};
+    ///
+    /// // Starts a server for localhost on port 8080
+    /// let mut server: Server = Server::new("localhost", 8080);
+    ///
+    /// // Define a route
+    /// server.route(Method::GET, "/", |req| {
+    ///     Response::new(
+    ///         200,
+    ///         "N O S E",
+    ///         vec![Header::new("Content-Type", "text/plain")],
+    ///     )
+    /// });
+    ///
+    /// // Starts the server with 8 threads
+    /// // This is blocking
+    /// // Keep the server from starting and blocking the main thread
+    /// # server.set_run(false);
+    /// server.start_threaded(8);
+    /// ```
+    #[cfg(feature = "thread_pool")]
+    pub fn start_threaded(&self, threads: usize) {
+        // Exit if the server should not run
+        if !self.run {
+            return;
         }
 
-        let stream_string = str::from_utf8(&buffer).expect("Error parsing buffer data");
+        let listener = init_listener(self.ip, self.port).unwrap();
+        let pool = ThreadPool::new(threads);
 
-        // Make Request Object
-        let req_method = get_request_method(stream_string.to_string());
-        let req_path = get_request_path(stream_string.to_string());
-        let req_query = get_request_query(stream_string.to_string());
-        let body = get_request_body(stream_string.to_string());
-        let headers = get_request_headers(stream_string.to_string());
-        let req = Request::new(
-            req_method,
-            &req_path,
-            req_query,
-            headers,
-            body,
-            stream.peer_addr().unwrap().to_string(),
-            stream_string.to_string(),
-        );
+        for event in listener.incoming() {
+            // Read stream into buffer
+            let stream = event.unwrap();
 
-        // Use middleware to handle request
-        // If middleware returns a `None`, the request will be handled by earlier middleware then the routes
-        for middleware in self.middleware.iter().rev() {
-            match (middleware)(&req) {
-                None => (),
-                Some(res) => return res,
-            }
+            let routes = self.routes.clone();
+            // let middleware = self.middleware.clone();
+            let error_handler = self.error_handler;
+
+            pool.execute(move || {
+                let mut stream = stream;
+                // Get the response from the handler
+                // Uses the most recently defined route that matches the request
+                let mut res = handle_connection(&stream, &Vec::new(), error_handler, &routes);
+
+                // Add default headers to response
+                let mut headers = res.headers;
+                // headers.append(&mut self.default_headers.clone().unwrap_or_default());
+
+                // Add content-length header to response
+                headers.push(Header::new("Content-Length", &res.data.len().to_string()));
+
+                // Convert the response to a string
+
+                let mut response = format!(
+                    "HTTP/1.1 {} OK\r\n{}\r\n\r\n",
+                    res.status,
+                    headers_to_string(headers)
+                )
+                .as_bytes()
+                .to_vec();
+
+                // Add Bytes of data to response
+                response.append(&mut res.data);
+
+                // Send the response
+                stream.write_all(&response).unwrap();
+                stream.flush().unwrap();
+            });
         }
+    }
 
-        // Loop through all routes and check if the request matches
-        for route in self.routes.iter().rev() {
-            if (req.method == route.method || route.method == Method::ANY)
-                && (req.path == route.path || route.path == "*")
-            {
-                return (route.handler)(req);
-            }
-        }
-
-        // If no route was found, return a default 404
-        Response::new(
-            404,
-            "Not Found",
-            vec![Header::new("Content-Type", "text/plain")],
-        )
+    fn handle_connection(&self, stream: &TcpStream) -> Response {
+        handle_connection(stream, &self.middleware, self.error_handler, &self.routes)
     }
 
     /// Keep a server from starting
@@ -266,6 +282,35 @@ impl Server {
         self.run = run;
     }
 
+    #[cfg(feature = "panic_handler")]
+    /// Set the panic handler response
+    ///
+    /// Default response is 500 "Internal Server Error :/"
+    ///
+    /// This is only available if the `panic_handler` feature is enabled
+    ///
+    /// Make sure that this wont panic because then the thread will crash
+    /// ## Example
+    /// ```rust
+    /// // Import Library
+    /// use afire::{Server, Response};
+    ///
+    /// // Create a server for localhost on port 8080
+    /// let mut server: Server = Server::new("localhost", 8080);
+    ///
+    /// // Set the panic handler response
+    /// server.set_error_handler(|_req| {
+    ///    Response::new(500, "Internal Server Error", vec![])
+    /// });
+    ///
+    /// // Start the server
+    /// # server.set_run(false);
+    /// server.start();
+    /// ```
+    pub fn set_error_handler(&mut self, res: fn(Request) -> Response) {
+        self.error_handler = res;
+    }
+
     /// Get the ip a server is listening on as a string
     ///
     /// For example, "127.0.0.1"
@@ -377,6 +422,7 @@ impl Server {
     /// server.start();
     /// ```
     /// Now you can make any type of request to `/nose` and it will return a 200
+    #[deprecated(since = "0.1.5", note = "Instead use .route(Method::ANY...)")]
     pub fn any(&mut self, path: &str, handler: fn(Request) -> Response) {
         self.routes
             .push(Route::new(Method::ANY, path.to_string(), handler));
@@ -442,6 +488,110 @@ impl Server {
     }
 }
 
+/// Handle a request
+fn handle_connection(
+    mut stream: &TcpStream,
+    middleware: &[Box<dyn Fn(&Request) -> Option<Response>>],
+    error_handler: fn(Request) -> Response,
+    routes: &[Route],
+) -> Response {
+    // Init (first) Buffer
+    let mut buffer = vec![0; BUFF_SIZE];
+
+    // Read stream into buffer
+    let _ = stream.read(&mut buffer).unwrap();
+
+    // Get buffer as string
+    let buffer_clone = buffer.clone();
+    let stream_string = match str::from_utf8(&buffer_clone) {
+        Ok(s) => s,
+        Err(_) => return Response::new(500, "Internal Server Error", vec![]),
+    };
+
+    // Get Content-Length header
+    // If header shows thar more space is needed,
+    // make a new buffer read the rest of the stream and add it to the first buffer
+    for i in get_request_headers(stream_string.to_string()) {
+        if i.name != "Content-Length" {
+            continue;
+        }
+        let header_size = get_header_size(stream_string.to_string());
+        let content_length = i.value.parse::<usize>().unwrap_or(0);
+        let new_buffer_size = content_length as i64 + header_size as i64 - BUFF_SIZE as i64;
+        if new_buffer_size > 0 {
+            let mut new_buffer = vec![0; new_buffer_size as usize];
+            let _ = stream.read(&mut new_buffer).unwrap();
+            buffer.append(&mut new_buffer);
+        }
+        break;
+    }
+
+    let stream_string = str::from_utf8(&buffer).expect("Error parsing buffer data");
+
+    // Make Request Object
+    let req_method = get_request_method(stream_string.to_string());
+    let req_path = get_request_path(stream_string.to_string());
+    let req_query = get_request_query(stream_string.to_string());
+    let body = get_request_body(stream_string.to_string());
+    let headers = get_request_headers(stream_string.to_string());
+    #[cfg(feature = "cookies")]
+    let cookies = get_request_cookies(stream_string.to_string());
+    let req = Request::new(
+        req_method,
+        &req_path,
+        req_query,
+        headers,
+        #[cfg(feature = "cookies")]
+        cookies,
+        body,
+        stream.peer_addr().unwrap().to_string(),
+        stream_string.to_string(),
+    );
+
+    // Use middleware to handle request
+    // If middleware returns a `None`, the request will be handled by earlier middleware then the routes
+    for middleware in middleware.iter().rev() {
+        match (middleware)(&req) {
+            None => (),
+            Some(res) => return res,
+        }
+    }
+
+    // Loop through all routes and check if the request matches
+    for route in routes.iter().rev() {
+        if (req.method == route.method || route.method == Method::ANY)
+            && (req.path == route.path || route.path == "*")
+        {
+            // Optionally enable automatic panic handling
+            #[cfg(feature = "panic_handler")]
+            {
+                let result = panic::catch_unwind(|| (route.handler)(req.clone()));
+                return result.ok().unwrap_or_else(|| (error_handler)(req));
+            }
+
+            #[cfg(not(feature = "panic_handler"))]
+            {
+                return (route.handler)(req);
+            }
+        }
+    }
+
+    // If no route was found, return a default 404
+    Response::new(
+        404,
+        &format!("Cannot {} {}", req.method, req.path),
+        vec![Header::new("Content-Type", "text/plain")],
+    )
+}
+
+/// Init Listaner
+fn init_listener(ip: [u8; 4], port: u16) -> Result<TcpListener, io::Error> {
+    TcpListener::bind(SocketAddr::new(
+        IpAddr::V4(Ipv4Addr::new(ip[0], ip[1], ip[2], ip[3])),
+        port,
+    ))
+}
+
 /// Get the request method of a raw HTTP request.
 fn get_request_method(raw_data: String) -> Method {
     let method_str = raw_data
@@ -485,7 +635,7 @@ fn get_request_query(raw_data: String) -> Query {
         if path.len() <= 1 {
             return Query::new_empty();
         }
-        return Query::new(&path[1]);
+        return Query::new(path[1]);
     }
     Query::new_empty()
 }
@@ -515,6 +665,24 @@ fn get_request_headers(raw_data: String) -> Vec<Header> {
     headers
 }
 
+/// Get Cookies of a raw HTTP request.
+#[cfg(feature = "cookies")]
+pub fn get_request_cookies(raw_data: String) -> Vec<Cookie> {
+    let spilt = raw_data.split("\r\n\r\n").collect::<Vec<&str>>();
+    let raw_headers = spilt[0].split("\r\n").collect::<Vec<&str>>();
+
+    for header in raw_headers {
+        if !header.starts_with("Cookie:") {
+            continue;
+        }
+
+        if let Some(cookie) = Cookie::from_string(header.trim_matches(char::from(0))) {
+            return cookie;
+        }
+    }
+    Vec::new()
+}
+
 /// Get the byte size of the headers of a raw HTTP request.
 fn get_header_size(raw_data: String) -> usize {
     let headers = raw_data.split("\r\n\r\n").collect::<Vec<&str>>();
